#![allow(clippy::expect_fun_call)]
use core::ops::Range;

use web_sys::{HtmlInputElement, HtmlTableCellElement, wasm_bindgen::JsCast};
use yew::prelude::*;

struct BitsTable {
    bits: Html,
    value_box: Html,
}

const HEX: &[char] = &[
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f', 'A', 'B', 'C',
    'D', 'E', 'F', 'x',
];

const ALLOWED_CHARS: &[char] = &[
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f', 'A', 'B', 'C',
    'D', 'E', 'F', 'x', '_',
];

fn value_to_u32(value: &str) -> u32 {
    let value = value.trim();
    if let Some(value) = value.strip_prefix("0x") {
        u32::from_str_radix(value.trim(), 16)
    } else if let Some(value) = value.strip_prefix("0b") {
        u32::from_str_radix(value.trim(), 2)
    } else if let Some(value) = value.strip_prefix("0o") {
        u32::from_str_radix(value.trim(), 8)
    } else {
        value.parse()
    }
    .unwrap_or_default()
}

impl Component for BitsTable {
    type Message = ();

    type Properties = ();

    fn create(_ctx: &Context<Self>) -> Self {
        let bit_click = |bit: u8| {
            move |ev: MouseEvent| {
                let input = web_sys::window()
                    .unwrap()
                    .document()
                    .unwrap()
                    .get_element_by_id("bytes_in")
                    .unwrap()
                    .dyn_into::<HtmlInputElement>()
                    .unwrap();
                let mut value = value_to_u32(&input.value());
                if let Some(target) = ev
                    .target()
                    .and_then(|t| t.dyn_into::<HtmlTableCellElement>().ok())
                {
                    match target.inner_text().as_str() {
                        "0" => {
                            target.set_inner_text("1");
                            value |= 1 << bit;
                        }
                        _ => {
                            target.set_inner_text("0");
                            value &= !(1 << bit);
                        }
                    }
                }

                input.set_value(format!("{value:#X}").as_str());

                input
                    .dispatch_event(&InputEvent::new("input").unwrap())
                    .unwrap();
            }
        };
        let bits = (0..4usize)
            .rev()
            .map(|idx| {
                let headers = (idx * 8..idx * 8 + 8)
                    .rev()
                    .map(|idx| {
                        html! {
                            <th>{idx}</th>
                        }
                    })
                    .collect::<Html>();
                let values = (idx * 8..idx * 8 + 8)
                    .rev()
                    .map(|idx| html! {
						<th class="bit" id={format!("bit{idx}")} onclick={bit_click(idx as _)}>{0}</th>
					})
                    .collect::<Html>();

                html! {
                    <table>
                        <tr class="header">
                            {headers}
                        </tr>
                        <tr>
                            {values}
                        </tr>
                    </table>
                }
            })
            .collect::<Html>();

        let on_input = move |ev: InputEvent| {
            let input_element = ev
                .target()
                .and_then(|t| t.dyn_into::<HtmlInputElement>().ok())
                .expect("on_input should have a target");
            let text = input_element.value().matches(HEX).collect::<String>();
            input_element.set_value(
                &input_element
                    .value()
                    .matches(ALLOWED_CHARS)
                    .collect::<String>(),
            );
            let mut num: u32 = match text.strip_prefix("0x") {
                Some(hex) => u32::from_str_radix(hex, 16).unwrap_or_default(),
                None => text.parse().unwrap_or_default(),
            };
            let doc = web_sys::window().unwrap().document().unwrap();
            if num == 0 {
                doc.get_element_by_id("activated_wrap")
                    .unwrap()
                    .set_attribute("style", "visibility: hidden;")
                    .unwrap();
            } else {
                doc.get_element_by_id("activated_wrap")
                    .unwrap()
                    .set_attribute("style", "")
                    .unwrap();
            }
            let mut act: Vec<Range<u8>> = vec![];
            for idx in 0..32u8 {
                let cell = doc
                    .get_element_by_id(format!("bit{idx}").as_str())
                    .expect("missing bit")
                    .dyn_into::<HtmlTableCellElement>()
                    .expect(format!("bit{idx} should be a table cell").as_str());
                let set = match num & 1 {
                    0 => false,
                    1 => true,
                    _ => unreachable!(),
                };
                if set {
                    match act.iter_mut().last() {
                        Some(last) => {
                            if last.end + 1 == idx {
                                last.end += 1;
                            } else {
                                act.push(idx..idx);
                            }
                        }
                        None => act.push(idx..idx),
                    }
                }
                cell.set_inner_text(match set {
                    false => "0",
                    true => "1",
                });
                num >>= 1;
            }
            if act.is_empty() {
                doc.get_element_by_id("activated")
                    .unwrap()
                    .set_text_content(None);
                return;
            }

            let act_str = act
                .into_iter()
                .map(|act| {
                    if act.is_empty() {
                        act.start.to_string()
                    } else {
                        format!("{act:?}")
                    }
                })
                .collect::<Vec<_>>()
                .join(", ");

            doc.get_element_by_id("activated")
                .unwrap()
                .set_text_content(Some(&act_str));
        };

        let value_box = html! {
            <div id="input_box">
                <input id="bytes_in" oninput={on_input} />
            </div>
        };
        Self { bits, value_box }
    }

    fn view(&self, _ctx: &Context<Self>) -> Html {
        let bits = self.bits.clone();
        let value_box = self.value_box.clone();

        html! {
            <>
            {value_box}
            <br />
            <div id="bit_wrap">
                {bits}
            </div>
            <div id="activated_ext">
                <h2>{"activated"}</h2>
            </div>
            <div id="activated_wrap">
                <p id="activated"></p>
            </div>
            </>
        }
    }
}

fn main() {
    wasm_logger::init(wasm_logger::Config::default());

    yew::Renderer::<BitsTable>::new().render();
}
